import os
import sys

from pandaclient import pcontainer_core

os.environ['PANDA_EXEC_STRING'] = 'pcontainer'

optP = pcontainer_core.make_arg_parse()

options = optP.parse_args()

status, output = pcontainer_core.submit(options)
sys.exit(status)



import os
import json
import argparse
import tempfile
import subprocess

# make arg parse with option definitions
def make_arg_parse():
    usage = 'pcontainer [options]\n' \
            '  HowTo is available at https://twiki.cern.ch/twiki/bin/view/PanDA/PandaContainer'
    optP = argparse.ArgumentParser(usage=usage, conflict_handler="resolve")
    optP.add_argument('--version',action='store_const', const=True, dest='version', default=None,
                      help='Displays version')
    optP.add_argument('--loadJson', action='store', dest='loadJson', default=None,
                      help='Read command-line parameters from a json file which contains a dict of {parameter: value}')
    optP.add_argument('--dumpJson', action='store', dest='dumpJson', default=None,
                      help='Dump all command-line parameters and submission result such as returnCode, returnOut, ' 
                           'jediTaskID, etc to a json file')
    optP.add_argument('--cvmfs', action='store_const', const=True, dest='cvmfs', default=False,
                      help="Bind /cvmfs to the container, bool, default False")
    optP.add_argument('--noX509', action='store_const', const=True, dest='noX509', default=False,
                      help="Unset X509 environment in the container, bool, default False")
    optP.add_argument('--datadir', action='store', dest='datadir', default='',
                      help="Binds the job directory to datadir for I/O operations, string, default /ctrdata")
    optP.add_argument('--workdir', action='store', dest='workdir', default='',
                      help="chdir to workdir in the container, string, default /ctrdata")
    optP.add_argument('--debug', action='store_const', const=True, dest='debug', default=False,
                      help="Enable more verbose output from runcontainer, bool, default False")
    optP.add_argument('--containerImage', action='store', dest='containerImage', default=None,
                      help="Name of a container image")
    optP.add_argument('--useCentralRegistry', action='store_const', const=True,
                         dest='useCentralRegistry', default=False,
                         help="Use the central container registry when --containerImage is used")
    optP.add_argument('--excludedSite', action='append', dest='excludedSite', default=None,
                      help="list of sites which are not used for site section, e.g., ANALY_ABC,ANALY_XYZ")
    optP.add_argument('--site', action='store', dest='site', default=None,
                      help='Site name where jobs are sent. If omitted, jobs are automatically sent to sites '
                           'where input is available. A comma-separated list of sites can be specified '
                           '(e.g. siteA,siteB,siteC), so that best sites are chosen from the given site list')
    optP.add_argument('--architecture', action='store', dest='architecture', default='',
                      help="CPU and/or GPU requirements. #CPU_spec&GPU_spec where CPU or GPU spec can be "
                           "omitted. CPU_spec = architecture<-vendor<-instruction set>>, "
                           "GPU_spec = vendor<-model>. A wildcards can be used if there is no special "
                           "requirement for the attribute. E.g., #x86_64-*-avx2&nvidia to ask for x86_64 "
                           "CPU with avx2 support and nvidia GPU")
    optP.add_argument('--noSubmit', action='store_const', const=True, dest='noSubmit', default=None,
                      help=argparse.SUPPRESS)
    optP.add_argument('--outDS', action='store', dest='outDS', default=None,
                      help='Base name of the output dataset container. Actual output dataset name is defined '
                           'for each output file type')
    optP.add_argument('--outputs', action='store', dest='outputs', default=None,
                      help='Output file names. Comma separated. e.g., --outputs out1.dat,out2.txt. You can specify '
                           'a suffix for each output container like <datasetNameSuffix>:<outputFileName>. '
                           'e.g., --outputs AAA:out1.dat,BBB:out2.txt, so that output container names are outDS_AAA/ '
                           'and outDS_BBB/ instead of outDS_out1.dat/ and outDS_out2.txt/')
    optP.add_argument('--intrSrv', action='store_const', const=True, dest='intrSrv', default=None,
                      help=argparse.SUPPRESS)
    optP.add_argument('--exec', action='store', dest='exec', default=None,
                      help='execution string. e.g., --exec "./myscript arg1 arg2"')
    optP.add_argument('-v', '--verbose', action='store_const', const=True, dest='verbose', default=None,
                      help='Verbose')
    optP.add_argument('--priority', action='store', dest='priority', default=None, type=int,
                      help='Set priority of the task (1000 by default). The value must be between 900 and 1100. ' \
                           'Note that priorities of tasks are relevant only in ' \
                           "each user's share, i.e., your tasks cannot jump over other user's tasks " \
                           'even if you give higher priorities.')
    optP.add_argument('--useSandbox', action='store_const', const=True, dest='useSandbox', default=False,
                      help='To send files in the run directory to remote sites which are not sent out by default ' \
                           'when --containerImage is used')
    optP.add_argument("-3", action="store_true", dest="python3", default=False,
                      help="Use python3")

    return optP

# construct command-line options from arg parse
def construct_cli_options(options):
    options = vars(options)
    if 'loadJson' in options and options['loadJson'] is not None:
        newOpts = json.load(open(options['loadJson']))
    else:
        newOpts = dict()
    for key in options:
        val = options[key]
        if key in ['loadJson']:
            continue
        if key == 'architecture':
            key = 'cmtConfig'
        if key == 'cvmfs':
            key = 'ctrCvmfs'
        if key == 'noX509':
            key = 'ctrNoX509'
        if key == 'datadir':
            key = 'ctrDatadir'
        if key == 'workdir':
            key = 'ctrWorkdir'
        if key == 'debug':
            key = 'ctrDebug'
        if val is None:
            continue
        newOpts[key] = val
    newOpts['noBuild'] = True
    tmpLoadJson = tempfile.NamedTemporaryFile(delete=False, mode='w')
    json.dump(newOpts, tmpLoadJson)
    tmpLoadJson.close()
    return tmpLoadJson.name

# submit
def submit(options):
    tmpDumpJson = None
    if 'dumpJson' not in options:
        tmpDumpJson = tempfile.mkstemp()[-1]
        options['dumpJson'] = tmpDumpJson
    tmpLoadJson = construct_cli_options(options)
    com = ['prun']
    com += ['--loadJson={0}'.format(tmpLoadJson)]
    ret_val = subprocess.call(com)
    ret_dict = None
    if ret_val == 0:
        try:
            ret_dict = json.load(open(options['dumpJson']))
            if tmpDumpJson is not None:
                os.remove(tmpDumpJson)
        except Exception:
            pass
    os.remove(tmpLoadJson)
    return (ret_val, ret_dict)



import os
import re
try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

sectionName = 'book'
confFile = os.path.expanduser('%s/panda.cfg' % os.environ['PANDA_CONFIG_ROOT'])


# create config or add section when missing
parser=ConfigParser.ConfigParser()
newFlag = False
if not os.path.exists(confFile):
    # create new config
    newFlag = True
    # make dir
    try:
        os.makedirs(os.environ['PANDA_CONFIG_ROOT'])
    except Exception:
        pass
else:
    # add section
    parser.read(confFile)
    if not parser.has_section(sectionName):
        newFlag = True
# new file or missing section
if newFlag:
    # add section
    parser.add_section(sectionName)
    # set dummy time
    parser.set(sectionName,'last_synctime','')
    # keep old config just in case
    try:
        os.rename(confFile, '%s.back' % confFile)
    except Exception:
        pass
    # write
    confFH = open(confFile,'w')
    parser.write(confFH)
    confFH.close()


# get config
def getConfig():
    # instantiate parser
    parser=ConfigParser.ConfigParser()
    parser.read(confFile)
    # config class
    class _bookConfig:
        pass
    bookConf = _bookConfig()
    # expand sequencer section
    for key,val in parser.items(sectionName):
        # convert int/bool
        if re.search('^\d+$',val) is not None:
            val = int(val)
        elif re.search('true',val,re.I) is not None:
            val = True
        elif re.search('false',val,re.I) is not None:
            val = False
        # set attributes    
        setattr(bookConf,key,val)
    # return
    return bookConf


# update
def updateConfig(bookConf):
    # instantiate parser
    parser=ConfigParser.ConfigParser()
    parser.read(confFile)
    # set new values
    for attr in dir(bookConf):
        if not attr.startswith('_'):
            val = getattr(bookConf,attr)
            if val is not None:
                parser.set(sectionName,attr,val)
    # keep old config
    try:
        os.rename(confFile, '%s.back' % confFile)
    except Exception as e:
        print("WARNING : cannot make backup for %s with %s" % (confFile, str(e)))
        return
    # update conf
    conFH = open(confFile,'w')
    parser.write(conFH)
    # close
    conFH.close()



import uuid

inFileList = ['file1','file2','file3']

logDatasetName = 'panda.jeditest.log.{0}'.format(uuid.uuid4())

taskParamMap = {}

taskParamMap['nFilesPerJob'] = 1
taskParamMap['nFiles'] = len(inFileList)
#taskParamMap['nEventsPerInputFile']  = 10000
#taskParamMap['nEventsPerJob'] = 10000
#taskParamMap['nEvents'] = 25000
taskParamMap['noInput'] = True
taskParamMap['pfnList'] = inFileList
#taskParamMap['mergeOutput'] = True
taskParamMap['taskName'] = str(uuid.uuid4())
taskParamMap['userName'] = 'someone'
taskParamMap['vo'] = 'wlcg'
taskParamMap['taskPriority'] = 900
#taskParamMap['reqID'] = reqIdx
taskParamMap['architecture'] = 'power9'
taskParamMap['transUses'] = 'A'
taskParamMap['transHome'] = 'B'
taskParamMap['transPath'] = 'executable'
taskParamMap['processingType'] = 'step1'
taskParamMap['prodSourceLabel'] = 'test'
taskParamMap['taskType'] = 'test'
taskParamMap['workingGroup'] = 'groupA'
#taskParamMap['coreCount'] = 1
#taskParamMap['walltime'] = 1
taskParamMap['cloud'] = 'NA'
taskParamMap['site'] = 'TEST_PQ'
taskParamMap['log'] = {'dataset': logDatasetName,
                       'type':'template',
                       'param_type':'log',
                       'token':'local',
                       'destination':'local',
                       'value':'{0}.${{SN}}.log.tgz'.format(logDatasetName)}
outDatasetName = 'panda.jeditest.{0}'.format(uuid.uuid4())


taskParamMap['jobParameters'] = [
    {'type':'constant',
     'value':'-i "${IN/T}"',
     },
    {'type':'constant',
     'value': 'ecmEnergy=8000 runNumber=12345'
     },
    {'type':'template',
     'param_type':'output',
     'token':'local',     
     'destination':'local',
     'value':'outputEVNTFile={0}.${{SN}}.root'.format(outDatasetName),
     'dataset':outDatasetName,
     'offset':1000,
     },
    {'type':'constant',
     'value':'aaaa',
     },
    ]



import os
import sys
import ssl
import uuid
import json
import time
import glob
import base64
import datetime

try:
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError
except ImportError:
    from urllib.request import urlopen, Request
    from urllib.parse import urlencode
    from urllib.error import HTTPError
    raw_input = input


TOKEN_BASENAME = '.token'
CACHE_PREFIX = '.page_cache_'


# decode ID token
def decode_id_token(enc):
    enc = enc.split('.')[1]
    enc += '=' * (-len(enc) % 4)
    dec = json.loads(base64.urlsafe_b64decode(enc.encode()))
    return dec


# utility class
class OpenIdConnect_Utils:

    # constructor
    def __init__(self, auth_config_url, token_dir=None, log_stream=None, verbose=False):
        self.auth_config_url = auth_config_url
        if token_dir is None:
            token_dir = os.environ['PANDA_CONFIG_ROOT']
        self.token_dir = os.path.expanduser(token_dir)
        if not os.path.exists(token_dir):
            os.makedirs(token_dir)
        self.log_stream = log_stream
        self.verbose = verbose

    # get token path
    def get_token_path(self):
        return os.path.join(self.token_dir, TOKEN_BASENAME)

    # get device code
    def get_device_code(self, device_auth_endpoint, client_id, audience):
        if self.verbose:
            self.log_stream.debug('getting device code')
        data = {'client_id': client_id,
                'scope': "openid profile email offline_access ",#iam",
                'audience': audience}
        rdata = urlencode(data).encode()
        req = Request(device_auth_endpoint, rdata)
        req.add_header('content-type', 'application/x-www-form-urlencoded')
        try:
            conn = urlopen(req)
            text = conn.read()
            if self.verbose:
                self.log_stream.debug(text)
            return True, json.loads(text)
        except HTTPError as e:
            return False, 'code={0}. reason={1}. description={2}'.format(e.code, e.reason, e.read())
        except Exception as e:
            return False, str(e)

    # get ID token
    def get_id_token(self, token_endpoint, client_id, client_secret, device_code, interval, expires_in):
        self.log_stream.info('Ready to get ID token?')
        while True:
            sys.stdout.write("[y/n] \n")
            choice = raw_input().lower()
            if choice == 'y':
                break
            elif choice == 'n':
                return False, "aborted"
        if self.verbose:
            self.log_stream.debug('getting ID token')
        startTime = datetime.datetime.utcnow()
        data = {'client_id': client_id,
                'client_secret': client_secret,
                'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
                'device_code': device_code}
        rdata = urlencode(data).encode()
        req = Request(token_endpoint, rdata)
        req.add_header('content-type', 'application/x-www-form-urlencoded')
        while datetime.datetime.utcnow() - startTime < datetime.timedelta(seconds=expires_in):
            try:
                conn = urlopen(req)
                text = conn.read().decode()
                if self.verbose:
                    self.log_stream.debug(text)
                id_token = json.loads(text)['id_token']
                with open(self.get_token_path(), 'w') as f:
                    f.write(text)
                return True, id_token
            except HTTPError as e:
                text = e.read()
                try:
                    description = json.loads(text)
                    # pending
                    if description['error'] == "authorization_pending":
                        time.sleep(interval + 1)
                        continue
                except Exception:
                    pass
                return False, 'code={0}. reason={1}. description={2}'.format(e.code, e.reason, text)
            except Exception as e:
                return False, str(e)

    # refresh token
    def refresh_token(self, token_endpoint, client_id, client_secret, refresh_token_string):
        if self.verbose:
            self.log_stream.debug('refreshing token')
        data = {'client_id': client_id,
                'client_secret': client_secret,
                'grant_type': 'refresh_token',
                'refresh_token': refresh_token_string}
        rdata = urlencode(data).encode()
        req = Request(token_endpoint, rdata)
        req.add_header('content-type', 'application/x-www-form-urlencoded')
        try:
            conn = urlopen(req)
            text = conn.read()
            if self.verbose:
                self.log_stream.debug(text)
            id_token = json.loads(text)['id_token']
            with open(self.get_token_path(), 'w') as f:
                f.write(text)
            return True, id_token
        except HTTPError as e:
            return False, 'code={0}. reason={1}. description={2}'.format(e.code, e.reason, e.read())
        except Exception as e:
            return False, str(e)

    # fetch page
    def fetch_page(self, url):
        path = os.path.join(self.token_dir, CACHE_PREFIX + str(uuid.uuid5(uuid.NAMESPACE_URL, str(url))))
        if os.path.exists(path) and \
                datetime.datetime.now() - datetime.datetime.fromtimestamp(os.path.getmtime(path)) < \
                datetime.timedelta(hours=1):
            try:
                with open(path) as f:
                    return True, json.load(f)
            except Exception as e:
                self.log_stream.debug('cached {0} is corrupted: {1}'.format(os.path.basename(url), str(e)))
        if self.verbose:
            self.log_stream.debug('fetching {0}'.format(url))
        try:
            context = ssl._create_unverified_context()
            conn = urlopen(url, context=context)
            text = conn.read().decode()
            if self.verbose:
                self.log_stream.debug(text)
            with open(path, 'w') as f:
                f.write(text)
            with open(path) as f:
                return True, json.load(f)
        except HTTPError as e:
            return False, 'code={0}. reason={1}. description={2}'.format(e.code, e.reason, e.read())
        except Exception as e:
            return False, str(e)

    # check token expiry
    def check_token(self):
        token_file = self.get_token_path()
        if os.path.exists(token_file):
            with open(token_file) as f:
                if self.verbose:
                    self.log_stream.debug('check {0}'.format(token_file))
                try:
                    # decode ID token
                    data = json.load(f)
                    dec = decode_id_token(data['id_token'])
                    exp_time = datetime.datetime.utcfromtimestamp(dec['exp'])
                    delta = exp_time - datetime.datetime.utcnow()
                    if self.verbose:
                        self.log_stream.debug('token expiration time : {0} UTC'.\
                                              format(exp_time.strftime("%Y-%m-%d %H:%M:%S")))
                    # check expiration time
                    if delta < datetime.timedelta(minutes=5):
                        # return refresh token
                        if 'refresh_token' in data:
                            if self.verbose:
                                self.log_stream.debug('to refresh token')
                            return False, data['refresh_token'], dec
                    else:
                        # return valid token
                        if self.verbose:
                            self.log_stream.debug('valid token is available')
                        return True, data['id_token'], dec
                except Exception as e:
                    self.log_stream.error('failed to decode cached token with {0}'.format(e))
        if self.verbose:
            self.log_stream.debug('cached token unavailable')
        return False, None, None

    # run device authorization flow
    def run_device_authorization_flow(self):
        # check toke expiry
        s, o, dec = self.check_token()
        if s:
            # still valid
            return True, o
        refresh_token_string = o
        # get auth config
        s, o = self.fetch_page(self.auth_config_url)
        if not s:
            return False, "Failed to get Auth configuration: " + o
        auth_config = o
        # get endpoint config
        s, o = self.fetch_page(auth_config['oidc_config_url'])
        if not s:
            return False, "Failed to get endpoint configuration: " + o
        endpoint_config = o
        # refresh token
        if refresh_token_string is not None:
            s, o = self.refresh_token(endpoint_config['token_endpoint'], auth_config['client_id'],
                                 auth_config['client_secret'], refresh_token_string)
            # refreshed
            if s:
                return True, o
            else:
                if self.verbose:
                    self.log_stream.debug('failed to refresh token: {0}'.format(o))
        # get device code
        s, o = self.get_device_code(endpoint_config['device_authorization_endpoint'], auth_config['client_id'],
                                    auth_config['audience'])
        if not s:
            return False, 'Failed to get device code: ' + o
        # get ID token
        self.log_stream.info(("Please go to {0} and sign in. "
                         "Waiting until authentication is completed").format(o['verification_uri_complete']))
        if 'interval' in o:
            interval = o['interval']
        else:
            interval = 5
        s, o = self.get_id_token(endpoint_config['token_endpoint'], auth_config['client_id'],
                                 auth_config['client_secret'], o['device_code'], interval, o['expires_in'])
        if not s:
            return False, "Failed to get ID token: " + o
        self.log_stream.info('All set')
        return True, o

    # cleanup
    def cleanup(self):
        for patt in [TOKEN_BASENAME, CACHE_PREFIX]:
            for f in glob.glob(os.path.join(self.token_dir, patt + '*')):
                os.remove(f)



import json
import base64
import time
import copy
import sys
import re

try:
    long()
except Exception:
    long = int

from . import Client
from . import PLogger
from pandaclient import queryPandaMonUtils
from pandaclient import localSpecs
from pandaclient import PsubUtils


def is_reqid(id):
    """
    whether an id is a reqID (otherwise jediTaskID)
    """
    return (id < 10 ** 7)

def _get_one_task(self, taskID, verbose=False):
    """
    get one task spec by ID
    """
    ts, url, data = queryPandaMonUtils.query_tasks(username=self.username, jeditaskid=taskID,
                                                    verbose=verbose)
    if isinstance(data, list) and data:
        task = data[0]
        taskspec = localSpecs.LocalTaskSpec(task, source_url=url, timestamp=ts)
        return taskspec
    else:
        return None

def _get_tasks_from_reqid(self, reqID, verbose=False):
    """
    get a list of task spec by reqID
    """
    ts, url, data = queryPandaMonUtils.query_tasks(username=self.username, reqid=reqID,
                                                    verbose=verbose)
    if isinstance(data, list) and data:
        taskspec_list = []
        for task in data:
            taskspec = localSpecs.LocalTaskSpec(task, source_url=url, timestamp=ts)
            taskspec_list.append(taskspec)
        return taskspec_list
    else:
        return None


func_return_value = True


def check_task_owner(func):
    """
    sanity check decorator of user ownership vs the task
    """
    # Wrapper
    def wrapper(self, *args, **kwargs):
        # Make logger
        tmpLog = PLogger.getPandaLogger()
        # initialize
        ret = None
        # check proxy
        PsubUtils.check_proxy(self.verbose, None)
        # check task owner
        try:
            taskid = None
            if args:
                taskid = args[0]
            if taskid is None:
                tmpLog.error('no taskID sepcified, nothing done')
                return
            # taskspec = _get_one_task(self, taskid, self.verbose)
            if is_reqid(taskid):
                taskspec_list = _get_tasks_from_reqid(self, taskid, self.verbose)
            else:
                taskspec_list = [_get_one_task(self, taskid, self.verbose)]
        except Exception as e:
            tmpLog.error('got {0}: {1}'.format(e.__class__.__name__, e))
        else:
            ret = True
            if taskspec_list is None:
                sys.stdout.write('Permission denied: reqID={0} is not owned by {1} \n'.format(
                                    taskid, self.username))
                ret = False
            else:
                for taskspec in taskspec_list:
                    if taskspec is not None and taskspec.username == self.username:
                        args_new = (taskspec.jeditaskid,) + args[1:]
                        ret = ret and func(self, *args_new, **kwargs)
                    else:
                        sys.stdout.write('Permission denied: taskID={0} is not owned by {1} \n'.format(
                                            taskid, self.username))
                        ret = False
        global func_return_value
        func_return_value = ret
        return ret
    wrapper.original_func = func
    return wrapper


# core class for book keeping
class PBookCore(object):

    # constructor
    def __init__(self, verbose=False):
        # verbose
        self.verbose = verbose
        # check proxy
        PsubUtils.check_proxy(self.verbose, None)
        # user name
        username_from_proxy = PsubUtils.extract_voms_proxy_username()
        if username_from_proxy:
            self.username = username_from_proxy
            sys.stdout.write('PBook user: {0} \n'.format(self.username))
        else:
            sys.stderr.write('ERROR : Cannot get user name from proxy. Exit... \n')
            sys.exit(1)

    # kill
    @check_task_owner
    def kill(self, taskID):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # kill JEDI task
        tmpLog.info('Sending killTask command ...')
        status, output = Client.killTask(taskID, self.verbose)
        # communication error
        if status != 0:
            tmpLog.error(output)
            tmpLog.error('Failed to kill jediTaskID=%s' % taskID)
            return False
        tmpStat, tmpDiag = output
        if not tmpStat:
            tmpLog.error(tmpDiag)
            tmpLog.error('Failed to kill jediTaskID=%s' % taskID)
            return False
        tmpLog.info(tmpDiag)
        # done
        tmpLog.info('Done, jediTaskID=%s will be killed in 30 min' % taskID)
        return True

    # finish
    @check_task_owner
    def finish(self, taskID, soft=False):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # finish JEDI task
        tmpLog.info('Sending finishTask command ...')
        status, output = Client.finishTask(taskID, soft, self.verbose)
        # communication error
        if status != 0:
            tmpLog.error(output)
            tmpLog.error('Failed to finish jediTaskID=%s' % taskID)
            return False
        tmpStat, tmpDiag = output
        if not tmpStat:
            tmpLog.error(tmpDiag)
            tmpLog.error('Failed to finish jediTaskID=%s' % taskID)
            return False
        tmpLog.info(tmpDiag)
        # done
        tmpLog.info('Done, jediTaskID=%s will be finished soon' % taskID)
        return True

    # set debug mode
    def debug(self, pandaID, modeOn):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # check proxy
        PsubUtils.check_proxy(self.verbose, None)
        # set
        status,output = Client.setDebugMode(pandaID,modeOn,self.verbose)
        if status != 0:
            tmpLog.error(output)
            tmpLog.error('Failed to set debug mode for %s' % pandaID)
            return
        # done
        tmpLog.info(output)
        return

    # kill and retry
    def killAndRetry(self, taskID, newOpts=None):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # kill
        retK = self.kill(taskID)
        if not retK:
            return False
        # sleep
        tmpLog.info('Going to sleep for 3 sec')
        time.sleep(3)
        nTry = 6
        for iTry in range(nTry):
            # check if task terminated
            taskspec = _get_one_task(self, taskID, self.verbose)
            if taskspec is not None:
                if taskspec.is_terminated():
                    break
                else:
                    tmpLog.info('Some sub-jobs are still running')
            else:
                tmpLog.warning('Could not get task status from panda monitor...')
            if iTry + 1 < nTry:
                # sleep
                tmpLog.info('Going to sleep for 30 sec')
                time.sleep(30)
            else:
                tmpLog.info('Max attempts exceeded. Please try later')
                return False
        # retry
        return self.retry(taskID, newOpts=newOpts)


    # retry
    @check_task_owner
    def retry(self, taskID, newOpts=None):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # set an empty map since mutable default value is used
        if newOpts is None:
            newOpts = {}
        else:
            newOpts = copy.deepcopy(newOpts)
        # warning for PQ
        site = newOpts.get('site', None)
        excludedSite = newOpts.get('excludedSite', None)
        PsubUtils.get_warning_for_pq(site, excludedSite, tmpLog)
        # for JEDI
        status,out = Client.retryTask(  taskID,
                                        verbose=self.verbose,
                                        properErrorCode=True,
                                        newParams=newOpts)
        if status != 0:
            tmpLog.error(status)
            tmpLog.error(out)
            tmpLog.error('Failed to retry TaskID=%s' % taskID)
            return False
        tmpStat,tmpDiag = out
        if (tmpStat not in [0,True] and newOpts == {}) or (newOpts != {} and tmpStat != 3):
            tmpLog.error(tmpDiag)
            tmpLog.error('Failed to retry TaskID=%s' % taskID)
            return False
        tmpLog.info(tmpDiag)
        return True

    # recover lost files
    @check_task_owner
    def recover_lost_files(self, taskID, test_mode=False):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # kill JEDI task
        tmpLog.info('Sending recovery request ...')
        status, output = Client.send_file_recovery_request(taskID, test_mode, self.verbose)
        # communication error
        if status != 0:
            tmpLog.error(output)
            tmpLog.error('Communication failure with the server')
            return False
        tmpStat, tmpDiag = output
        if not tmpStat:
            tmpLog.error(tmpDiag)
            tmpLog.error('request was not received')
            return False
        # done
        tmpLog.info(tmpDiag)
        return True

    # get job metadata
    def getUserJobMetadata(self, taskID, output_filename):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # get metadata
        tmpLog.info('getting metadata')
        status, metadata = Client.getUserJobMetadata(taskID, verbose=self.verbose)
        if status != 0:
            tmpLog.error(metadata)
            tmpLog.error("Failed to get metadata")
            return False
        with open(output_filename, 'w') as f:
            json.dump(metadata, f)
        tmpLog.info('dumped to {0}'.format(output_filename))
        # return
        return True

    # get task specs of active tasks
    def get_active_tasks(self):
        """
        get all reachable task specs of the user
        """
        active_superstatus_str = '|'.join(localSpecs.task_active_superstatus_list)
        ts, url, data = queryPandaMonUtils.query_tasks(username=self.username, superstatus=active_superstatus_str,
                                                        verbose=self.verbose)
        if isinstance(data, list) and list:
            taskspec_list = [ localSpecs.LocalTaskSpec(task, source_url=url, timestamp=ts) for task in data ]
            return taskspec_list
        else:
            return None

    # show status
    def show(self, some_ids=None, username=None, limit=1000, taskname=None, days=14, jeditaskid=None,
                reqid=None, status=None, superstatus=None, metadata=False, sync=False, format='standard'):
        # user name
        if username is None:
            username = self.username
        # shortcut of jeditaskid and reqid
        if isinstance(some_ids, (int, long)):
            if is_reqid(some_ids):
                reqid = str(some_ids)
            else:
                jeditaskid = str(some_ids)
        elif isinstance(some_ids, (list, tuple)) and some_ids:
            first_id = some_ids[0]
            ids_str = '|'.join([str(x) for x in some_ids])
            if first_id and isinstance(first_id, (int, long)) and is_reqid(first_id):
                reqid = ids_str
            else:
                jeditaskid = ids_str
        elif some_ids == 'run':
            superstatus = '|'.join(localSpecs.task_active_superstatus_list)
        elif some_ids == 'fin':
            superstatus = '|'.join(localSpecs.task_final_superstatus_list)
        # print
        if format != 'json':
            sys.stderr.write('Showing only max {limit} tasks in last {days} days. One can set days=N to see tasks in last N days, and limit=M to see at most M latest tasks \n'
                             .format(days=days, limit=limit))
        # query
        ts, url, data = queryPandaMonUtils.query_tasks( username=username, limit=limit, reqid=reqid,
                                                        status=status, superstatus=superstatus,
                                                        taskname=taskname, days=days, jeditaskid=jeditaskid,
                                                        metadata=metadata, sync=sync, verbose=self.verbose)
        # print header row
        _tmpts = localSpecs.LocalTaskSpec
        if format in ['json', 'plain']:
            pass
        elif format == 'long':
            print(_tmpts.head_dict['long'])
        else:
            print(_tmpts.head_dict['standard'])
        # print tasks
        if format == 'json':
            return data
        elif format == 'plain':
            for task in data:
                taskspec = localSpecs.LocalTaskSpec(task, source_url=url, timestamp=ts)
                taskspec.print_plain()
        elif format == 'long':
            i_count = 1
            for task in data:
                taskspec = localSpecs.LocalTaskSpec(task, source_url=url, timestamp=ts)
                if i_count % 10 == 0:
                    print(_tmpts.head_dict['long'])
                taskspec.print_long()
                i_count += 1
        else:
            for task in data:
                taskspec = localSpecs.LocalTaskSpec(task, source_url=url, timestamp=ts)
                taskspec.print_standard()

    # execute workflow command
    def execute_workflow_command(self, command_name, request_id):
        tmpLog = PLogger.getPandaLogger()
        tmpLog.info('executing {}'.format(command_name))
        status, output = Client.call_idds_user_workflow_command(command_name, {'request_id': request_id},
                                                                self.verbose)
        if status != 0:
            tmpLog.error(output)
            tmpLog.error("Failed to execute {}".format(command_name))
            return False, None
        if not output[0]:
            tmpLog.error(output[-1])
            tmpLog.error("Failed to execute {}".format(command_name))
            return False, None
        return True, output[-1]

    # set secret
    def set_secret(self, key, value, is_file=False):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        if is_file:
            with open(value, 'rb') as f:
                value = base64.b64encode(f.read()).encode()
            # add prefix
            key = '___file___:'+format(key)
        size_limit = 1000
        if value and len(value) > size_limit * 1024:
            tmpLog.error("The value length exceeds the limit ({0} kB)".format(size_limit))
            return False
        status, output = Client.set_user_secert(key, value, verbose=self.verbose)
        if status != 0:
            tmpLog.error(output)
            tmpLog.error("Failed to set secret")
            return False
        status, msg = output
        if status:
            tmpLog.info(msg)
        else:
            tmpLog.error(msg)
        # return
        return status

    # list secrets
    def list_secrets(self, full=False):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        status, output = Client.get_user_secerts(verbose=self.verbose)
        if status != 0:
            tmpLog.error(output)
            tmpLog.error("Failed to get secrets")
            return False
        status, data = output
        if status:
            if data:
                prefix = '^___[a-z]+___:'
                msg = '\n'
                keys = [re.sub(prefix, '', k) for k in data.keys()]
                big_key = len(max(keys, key=len))
                template = '{{:{}s}}: {{}}\n'.format(big_key+1)
                msg += template.format('Key', 'Value')
                msg += template.format('-'*big_key, '-'*20)
                keys.sort()
                max_len = 50
                for k in data:
                    value = data[k]
                    # hide prefix
                    if re.search(prefix, k):
                        k = re.sub(prefix, '', k)
                    if not full and len(value) > max_len:
                        value = value[:max_len] + '...'
                    msg += template.format(k, value)
            else:
                msg = "No secrets"
            print(msg)
        else:
            tmpLog.error(data)
        # return
        return status

    # pause
    def pause(self, task_id):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # pause JEDI task
        tmpLog.info('Sending pause command ...')
        status, output = Client.pauseTask(task_id, self.verbose)
        # communication error
        if status != 0:
            tmpLog.error(output)
            tmpLog.error('Failed to pause jediTaskID=%s' % task_id)
            return False
        tmpStat, tmpDiag = output
        if tmpStat != 0:
            print(tmpStat)
            tmpLog.error(tmpDiag)
            tmpLog.error('Failed to pause jediTaskID=%s' % task_id)
            return False
        tmpLog.info(tmpDiag)
        # done
        tmpLog.info('Done')
        return True

    # resume
    def resume(self, task_id):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # pause JEDI task
        tmpLog.info('Sending resume command ...')
        status, output = Client.resumeTask(task_id, self.verbose)
        # communication error
        if status != 0:
            tmpLog.error(output)
            tmpLog.error('Failed to resume jediTaskID=%s' % task_id)
            return False
        tmpStat, tmpDiag = output
        if tmpStat != 0:
            tmpLog.error(tmpDiag)
            tmpLog.error('Failed to resume jediTaskID=%s' % task_id)
            return False
        tmpLog.info(tmpDiag)
        # done
        tmpLog.info('Done')
        return True



import os
import sys
import json
import copy
import tempfile
import importlib

from . import PLogger
from . import Client
from . import PBookCore


class PandaAPI(object):

    # constructor
    def __init__(self):
        self.command_body = {}
        self.pbook = None
        self.log = PLogger.getPandaLogger()

    # kill a task
    def kill_task(self, task_id, verbose=True):
        """kill a task
           args:
              task_id: jediTaskID of the task to be killed
              verbose: True to see debug messages
           returns:
              status code
                 0: communication succeeded to the panda server
               255: communication failure
              tuple of return code and diagnostic message
                 0: request is registered
                 1: server error
                 2: task not found
                 3: permission denied
                 4: irrelevant task status
               100: non SSL connection
               101: irrelevant taskID
        """
        return Client.killTask(task_id)

    # finish a task
    def finish_task(self, task_id, wait_running=False, verbose=False):
        """finish a task
           args:
              task_id: jediTaskID of the task to finish
              wait_running: True to wait until running jobs are done
              verbose: True to see debug messages
           returns:
              status code
                 0: communication succeeded to the panda server
               255: communication failure
              tuple of return code and diagnostic message
                 0: request is registered
                 1: server error
                 2: task not found
                 3: permission denied
                 4: irrelevant task status
               100: non SSL connection
               101: irrelevant taskID
        """
        return Client.finishTask(task_id, wait_running, verbose)

    # retry a task
    def retry_task(self, task_id, new_parameters=None, verbose=False):
        """retry a task
           args:
              task_id: jediTaskID of the task to retry
              new_parameters: a dictionary of task parameters to overwrite
              verbose: True to see debug messages
           returns:
              status code
                 0: communication succeeded to the panda server
               255: communication failure
              tuple of return code and diagnostic message
                 0: request is registered
                 1: server error
                 2: task not found
                 3: permission denied
                 4: irrelevant task status
               100: non SSL connection
               101: irrelevant taskID
        """
        return Client.retryTask(task_id, verbose, True, new_parameters)

    # get tasks
    def get_tasks(self, task_ids=None, limit=1000, days=14, status=None, username=None):
        """get a list of task dictionaries
           args:
              task_ids: a list of task IDs, or None to get recent tasks
              limit: the max number of tasks to fetch from the server
              days: tasks for last N days to fetch
              status: filtering with task status
              username: user name of the tasks, or None to get own tasks
           returns:
              a list of task dictionaries
        """
        if not self.pbook:
            self.pbook = PBookCore.PBookCore()
        return self.pbook.show(task_ids, limit=limit, days=days, format='json', status=status,
                               username=username)

    # show tasks
    def show_tasks(self, task_ids=None, limit=1000, days=14, format='standard', status=None, username=None):
        """show tasks
           args:
              task_ids: a list of task IDs, or None to get recent tasks
              limit: the max number of tasks to fetch from the server
              days: tasks for last N days to fetch
              format: standard, long, or plain
              status: filtering with task status
              username: user name of the tasks, or None to get own tasks
           returns:
              None
        """
        if not self.pbook:
            self.pbook = PBookCore.PBookCore()
        self.pbook.show(task_ids, limit=limit, days=days, format=format, status=status, username=username)

    # submit a task
    def submit_task(self, task_params, verbose=False):
        """submit a task using low-level API
           args:
              task_params: a dictionary of task parameters
              verbose: True to see debug messages
           returns:
              status code
                 0: communication succeeded to the panda server
               255: communication failure
              tuple of return code, message from the server, and task ID if successful
                 0: request is processed
                 1: duplication in DEFT
                 2: duplication in JEDI
                 3: accepted for incremental execution
                 4: server error
        """
        return Client.insertTaskParams(task_params, verbose=verbose, properErrorCode=True)

    # get metadata of all jobs in a task
    def get_job_metadata(self, task_id, output_json_filename):
        """get metadata of all jobs in a task
           args:
              task_id: task ID
              output_json_filename: output json filename
        """
        if not self.pbook:
            self.pbook = PBookCore.PBookCore()
        return self.pbook.getUserJobMetadata(task_id, output_json_filename)

    # execute xyz
    def execute_xyz(self, command_name, module_name, args, console_log=True):
        dump_file = None
        stat = False
        ret = None
        err_str = None
        try:
            # convert args
            sys.argv = copy.copy(args)
            sys.argv.insert(0, command_name)
            # set dump file
            if '--dumpJson' not in sys.argv:
                sys.argv.append('--dumpJson')
                with tempfile.NamedTemporaryFile(delete=False) as f:
                    sys.argv.append(f.name)
                    dump_file = f.name
            # disable logging
            if not console_log:
                PLogger.disable_logging()
            # run
            if command_name in self.command_body:
                if hasattr(self.command_body[command_name], 'main'):
                    getattr(self.command_body[command_name], 'main')()
                else:
                    self.command_body[command_name].reload()
            else:
                self.command_body[command_name] = importlib.import_module(module_name)
                if hasattr(self.command_body[command_name], 'main'):
                    getattr(self.command_body[command_name], 'main')()
            stat = True
        except SystemExit as e:
            if e.code == 0:
                stat = True
            else:
                err_str = 'failed with code={0}'.format(e.code)
        except Exception as e:
            err_str = 'failed with {0}'.format(str(e))
        finally:
            # enable logging
            if not console_log:
                PLogger.enable_logging()
            if err_str:
                self.log.error(err_str)
            # read dump fle
            try:
                with open(sys.argv[sys.argv.index('--dumpJson') + 1]) as f:
                    ret = json.load(f)
                    if len(ret) == 1:
                        ret = ret[0]
            except Exception:
                pass
        # delete dump file
        if not dump_file:
            os.remove(dump_file)
        return stat, ret

    # execute prun
    def execute_prun(self, args, console_log=True):
        """execute prun command

           args:
               args: The arguments used to execute prun. This is a list of strings.
               console_log: False to disable console logging

           returns:
               status: True if succeeded. Otherwise, False
               a dictionary: Task submission attributes including jediTaskID
        """
        return self.execute_xyz('prun', 'pandaclient.PrunScript', args, console_log)

    # execute pathena
    def execute_pathena(self, args, console_log=True):
        """execute pathena command

           args:
               args: The arguments used to execute prun. This is a list of strings.
               console_log: False to disable console logging

           returns:
               status: True if succeeded. Otherwise, False
               a dictionary: Task submission attributes including jediTaskID
        """
        return self.execute_xyz('pathena', 'pandaclient.PathenaScript', args, console_log)

    # execute phpo
    def execute_phpo(self, args, console_log=True):
        """execute phpo command

           args:
               args: The arguments used to execute prun. This is a list of strings.
               console_log: False to disable console logging

           returns:
               status: True if succeeded. Otherwise, False
               a dictionary: Task submission attributes including jediTaskID
        """
        return self.execute_xyz('phpo', 'pandaclient.PhpoScript', args, console_log)

    # hello
    def hello(self, verbose=False):
        """Health check with the PanDA server
           args:
              verbose: True to see verbose message
           returns:
              status code
                 0: communication succeeded to the panda server
               255: communication failure
              diagnostic message
        """
        return Client.hello(verbose)

    # increase attempt numbers to retry failed jobs
    def increase_attempt_nr(self, task_id, increase=3, verbose=False):
        """increase attempt numbers to retry failed jobs
           args:
              task_id: jediTaskID of the task
              increase: increase for attempt numbers
              verbose: True to see verbose message
           returns:
              status code
                    0: communication succeeded to the panda server
                    255: communication failure
              return code
                    0: succeeded
                    1: unknown task
                    2: invalid task status
                    3: permission denied
                    4: wrong parameter
                    None: database error
        """
        return Client.increase_attempt_nr(task_id, increase, verbose)


pandaAPI = PandaAPI()
del PandaAPI


def get_api():
    return pandaAPI



import json
import sys
import re
import os
import copy
import atexit

from pandaclient.Group_argparse import GroupArgParser
from pandaclient import PLogger
from pandaclient import PandaToolsPkgInfo
from pandaclient import MiscUtils
from pandaclient import Client
from pandaclient import PsubUtils

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote

try:
    unicode
except Exception:
    unicode = str


# main
def main(get_taskparams=False, ext_args=None, dry_mode=False):

    # tweak sys.argv
    sys.argv.pop(0)
    sys.argv.insert(0, 'phpo')

    usage = """phpo [options]
    """

    optP = GroupArgParser(usage=usage, conflict_handler="resolve")

    group_input = optP.add_group('input', 'input dataset(s)/files/format')
    group_output = optP.add_group('output', 'output dataset/files')
    group_config = optP.add_group('config', 'single configuration file to set multiple options')
    group_submit = optP.add_group('submit', 'job submission/site/retry')
    group_expert = optP.add_group('expert', 'for experts/developers only')

    optP.add_helpGroup()

    group_config.add_argument('--version', action='store_const', const=True, dest='version', default=False,
                              help='Displays version')
    group_config.add_argument('--loadJson', action='store', dest='loadJson', default=None,
                              help='Read task parameters from a json file. Some parameters can be overridden '
                                   'by using command-line arguments')
    group_config.add_argument('--dumpJson', action='store', dest='dumpJson', default=None,
                              help='Dump all command-line parameters and submission result '
                                   'such as returnCode, returnOut, and jediTaskID to a json file')
    group_config.add_argument('--nParallelEvaluation', action='store', dest='nParallelEvaluation', default=1, type=int,
                              help='The number of hyperparameter points being evaluated concurrently. 1 by default')
    group_config.add_argument('--maxPoints', action='store', dest='maxPoints', default=10, type=int,
                              help='The max number of hyperparameter points to be evaluated in the entire search '
                                   '(for each segment in segmented HPO). '
                                   '10 by default')
    group_config.add_argument('--maxEvaluationJobs', action='store', dest='maxEvaluationJobs', default=None, type=int,
                              help='The max number of evaluation jobs in the entire search '
                                   '(for each segment in segmented HPO). 2*maxPoints by default. '
                                   'The task is terminated when all hyperparameter points are evaluated or '
                                   'the number of evaluation jobs reaches maxEvaluationJobs')
    group_config.add_argument('--maxPointsPerEvaluationJob', action='store', dest='maxPointsPerEvaluationJob',
                              default=None, type=int,
                              help='The max number of hyperparameter points taken in each evaluation job')
    group_config.add_argument('--nPointsPerIteration', action='store', dest='nPointsPerIteration', default=2, type=int,
                              help='The max number of hyperparameter points generated in each iteration. 2 by default '
                                   'Simply speaking, the steering container is executed maxPoints/nPointsPerIteration '
                                   'times when minUnevaluatedPoints is 0. The number of new points is '
                                   'nPointsPerIteration-minUnevaluatedPoints')
    group_config.add_argument('--minUnevaluatedPoints', action='store', dest='minUnevaluatedPoints', default=None, type=int,
                              help='The next iteration is triggered to generate new hyperparameter points when the number '
                                   'of unevaluated hyperparameter points goes below minUnevaluatedPoints. 0 by default')
    group_config.add_argument('--steeringContainer', action='store', dest='steeringContainer', default=None,
                              help='The container image for steering run by docker')
    group_config.add_argument('--steeringExec', action='store', dest='steeringExec', default=None,
                              help='Execution string for steering. If --steeringContainer is specified, the string '
                                   'is executed inside of the container. Otherwise, the string is used as command-line '
                                   'arguments for the docker command')
    group_config.add_argument('--searchSpaceFile', action='store', dest='searchSpaceFile', default=None,
                              help='External json filename to define the search space which is described as a dictionary. '
                                   'None by default. '
                                   'If this option is used together with --segmentSpecFile the json file contains a list '
                                   'of search space dictionaries. It is possible to contain only one search space '
                                   'dictionary if all segments use the same search space. In this case the search space '
                                   'dictionary is cloned for every segment')
    group_config.add_argument('--evaluationContainer', action='store', dest='evaluationContainer', default=None,
                              help='The container image for evaluation')
    group_config.add_argument('--evaluationExec', action='store', dest='evaluationExec', default=None,
                              help='Execution string to run evaluation in singularity')
    group_config.add_argument('--evaluationInput', action='store', dest='evaluationInput', default='input.json',
                              help='Input filename for evaluation where a json-formatted hyperparameter point is placed. '
                                   'input.json by default')
    group_config.add_argument('--evaluationTrainingData', action='store', dest='evaluationTrainingData',
                              default='input_ds.json',
                              help='Input filename for evaluation where a json-formatted list of training data filenames '
                                   'is placed. input_ds.json by default. Can be omitted if the payload directly fetches '
                                   'the training data using wget or something')
    group_config.add_argument('--evaluationOutput', action='store', dest='evaluationOutput', default='output.json',
                              help='Output filename of evaluation. output.json by default')
    group_config.add_argument('--evaluationMeta', action='store', dest='evaluationMeta', default=None,
                              help='The name of metadata file produced by evaluation')
    group_config.add_argument('--evaluationMetrics', action='store', dest='evaluationMetrics', default=None,
                              help='The name of metrics file produced by evaluation')
    group_config.add_argument('--checkPointToSave', action='store', dest='checkPointToSave', default=None,
                              help='A comma-separated list of files and/or directories to be periodically saved ' \
                                   'to a tarball for checkpointing. Note that those files and directories must be placed ' \
                                   'in the working directory. None by default')
    group_config.add_argument('--checkPointToLoad', action='store', dest='checkPointToLoad', default=None,
                              help='The name of the saved tarball for checkpointing. The tarball is given to ' \
                                   'the evaluation container when the training is resumed, if this option is specified. '
                                   'Otherwise, the tarball is automatically extracted in the working directories')
    group_config.add_argument('--checkPointInterval', action='store', dest='checkPointInterval', default=None, type=int,
                              help='Frequency to check files for checkpointing in minute. '
                                   '5 by default')
    group_config.add_argument('--alrbArgs', action='store', dest='alrbArgs', default=None,
                              help='Additional arguments for ALRB to run the evaluation container. ' \
                                   '"setupATLAS -c --help" shows available ALRB arguments. For example, ' \
                                   '--alrbArgs "--nocvmfs --nohome" to skip mounting /cvmfs and $HOME. ' \
                                   'This option is mainly for experts who know how the system and the container ' \
                                   'communicates with each other and how additional ALRB arguments affect ' \
                                   'the consequence')
    group_config.add_argument('--architecture', action='store', dest='architecture', default='',
                              help="CPU and/or GPU requirements. #CPU_spec&GPU_spec where CPU or GPU spec can be "
                                   "omitted. CPU_spec = architecture<-vendor<-instruction set>>, "
                                   "GPU_spec = vendor<-model>. A wildcards can be used if there is no special "
                                   "requirement for the attribute. E.g., #x86_64-*-avx2&nvidia to ask for x86_64 "
                                   "CPU with avx2 support and nvidia GPU")
    group_config.add_argument('--segmentSpecFile', action='store', dest='segmentSpecFile', default=None,
                              help='External json filename to define segments for segmented HPO which has one model '
                                   'for each segment to be optimized independently. The file '
                                   "contains a list of dictionaries {'name': arbitrary_unique_segment_name, "
                                   "'files': [filename_used_for_the_segment_in_the_training_dataset, ... ]}. "
                                   "It is possible to specify 'datasets' instead of 'files' in those dictionaries "
                                   "if the training dataset has constituent datasets and "
                                   "is partitioned with the constituent dataset boundaries. "
                                   'None by default')
    group_config.add_argument('-v', action='store_const', const=True, dest='verbose', default=False,
                              help='Verbose')

    group_input.add_argument('--trainingDS', action='store', dest='trainingDS', default=None,
                             help='Name of training dataset')

    group_output.add_argument('--outDS', action='store', dest='outDS', default=None,
                              help='Name of the dataset for output and log files')
    group_output.add_argument('--official', action='store_const', const=True, dest='official', default=False,
                              help='Produce official dataset')

    group_submit.add_argument('--site', action='store', dest='site', default=None,
                              help='The site name where jobs are sent. If omitted, jobs are automatically sent to sites '
                                   'where input is available. A comma-separated list of sites can be specified '
                                   '(e.g. siteA,siteB,siteC), so that best sites are chosen from the given site list')
    group_submit.add_argument('--workingGroup', action='store', dest='workingGroup', default=None,
                              help="set working group")
    group_submit.add_argument('--noSubmit', action='store_const', const=True, dest='noSubmit', default=False,
                              help="Dry-run")
    group_submit.add_argument("-3", action="store_true", dest="python3", default=False,
                              help="Use python3")
    group_submit.add_argument('--voms', action='store', dest='vomsRoles', default=None, type=str,
                              help="generate proxy with paticular roles. "
                                   "e.g., atlas:/atlas/ca/Role=production,atlas:/atlas/fr/Role=pilot")
    group_submit.add_argument('--noEmail', action='store_const', const=True, dest='noEmail', default=False,
                              help='Suppress email notification')

    group_expert.add_argument('--intrSrv', action='store_const', const=True, dest='intrSrv', default=False,
                              help="Please don't use this option. Only for developers to use the intr panda server")

    # get logger
    tmpLog = PLogger.getPandaLogger()

    options = optP.parse_args(ext_args)
    option_names = set(vars(options).keys())

    jsonExecStr = ''
    if options.loadJson is not None:
        with open(os.path.expanduser(options.loadJson)) as f:
            json_options = json.load(f)
            for k in json_options:
                if k in option_names:
                    v = json_options[k]
                    if isinstance(v, (str, unicode)):
                        try:
                            v = int(v)
                        except Exception:
                            pass
                    setattr(options, k, v)
                    if v is True:
                        jsonExecStr += ' --{0}'.format(k)
                    else:
                        if isinstance(v, (str, unicode)):
                            jsonExecStr += " --{0}='{1}'".format(k, v)
                        else:
                            jsonExecStr += " --{0}={1}".format(k, v)
                else:
                    tmpLog.warning('ignore unknown option {0} in {1}'.format(k, options.loadJson))

    if options.version:
        print("Version: %s" % PandaToolsPkgInfo.release_version)
        sys.exit(0)

    # check grid-proxy
    if not dry_mode:
        PsubUtils.check_proxy(options.verbose, options.vomsRoles)

    # check options
    # non_null_opts = ['outDS', 'evaluationContainer', 'evaluationExec', 'steeringContainer', 'steeringExec']
    non_null_opts = ['outDS', 'evaluationContainer', 'evaluationExec', 'steeringExec']
    for opt_name in non_null_opts:
        if getattr(options, opt_name) is None:
            tmpLog.error('--{0} is not specified'.format(opt_name))
            sys.exit(1)

    if not options.outDS.endswith('/'):
        options.outDS += '/'

    if options.maxEvaluationJobs is None:
        options.maxEvaluationJobs = 2 * options.maxPoints

    # check output name

    if not dry_mode:
        nickName = PsubUtils.getNickname()
        if not PsubUtils.checkOutDsName(options.outDS, options.official, nickName,
                                                     verbose=options.verbose):
            tmpStr = "invalid output dataset name: %s" % options.outDS
            tmpLog.error(tmpStr)
            sys.exit(1)

    # full execution string
    fullExecString = PsubUtils.convSysArgv(ext_args)
    fullExecString += jsonExecStr

    # use INTR server
    if options.intrSrv:
        Client.useIntrServer()

    # create tmp dir
    curDir = os.getcwd()
    tmpDir = os.path.join(curDir, MiscUtils.wrappedUuidGen())
    os.makedirs(tmpDir)


    # exit action
    def _onExit(dir, del_command):
        del_command('rm -rf %s' % dir)


    atexit.register(_onExit, tmpDir, MiscUtils.commands_get_output)

    # sandbox
    if options.verbose:
        tmpLog.debug("=== making sandbox ===")
    archiveName = 'jobO.%s.tar' % MiscUtils.wrappedUuidGen()
    archiveFullName = os.path.join(tmpDir, archiveName)
    if not dry_mode:
        extensions = ['json', 'py', 'sh', 'yaml']
        find_opt = ' -o '.join(['-name "*.{0}"'.format(e) for e in extensions])
        tmpOut = MiscUtils.commands_get_output('find . {0} | tar cvfz {1} --files-from - '.format(find_opt, archiveFullName))

        if options.verbose:
            print(tmpOut + '\n')
            tmpLog.debug("=== checking sandbox ===")
            tmpOut = MiscUtils.commands_get_output('tar tvfz {0}'.format(archiveFullName))
            print(tmpOut + '\n')

        if not options.noSubmit:
            if options.verbose:
                tmpLog.debug("=== uploading sandbox===")
            os.chdir(tmpDir)
            status, out = Client.putFile(archiveName, options.verbose, useCacheSrv=True, reuseSandbox=True)
            os.chdir(curDir)
            if out.startswith('NewFileName:'):
                # found the same input sandbox to reuse
                archiveName = out.split(':')[-1]
            elif out != 'True':
                # failed
                print(out)
                tmpLog.error("Failed with %s" % status)
                sys.exit(1)

    matchURL = re.search("(http.*://[^/]+)/", Client.baseURLCSRVSSL)
    sourceURL = matchURL.group(1)

    # making task params
    taskParamMap = {}

    taskParamMap['noInput'] = True
    taskParamMap['nEventsPerJob'] = 1
    taskParamMap['nEvents'] = options.nParallelEvaluation
    taskParamMap['maxNumJobs'] = options.maxEvaluationJobs
    taskParamMap['totNumJobs'] = options.maxPoints
    taskParamMap['taskName'] = options.outDS
    taskParamMap['vo'] = 'atlas'
    taskParamMap['architecture'] = options.architecture
    taskParamMap['hpoWorkflow'] = True
    taskParamMap['transUses'] = ''
    taskParamMap['transHome'] = ''
    taskParamMap['transPath'] = 'http://pandaserver.cern.ch:25080/trf/user/runHPO-00-00-01'
    taskParamMap['processingType'] = 'panda-client-{0}-jedi-hpo'.format(PandaToolsPkgInfo.release_version)
    taskParamMap['prodSourceLabel'] = 'user'
    taskParamMap['useLocalIO'] = 1
    taskParamMap['cliParams'] = fullExecString
    taskParamMap['skipScout'] = True
    if options.noEmail:
        taskParamMap['noEmail'] = True
    if options.workingGroup is not None:
        taskParamMap['workingGroup'] = options.workingGroup
    taskParamMap['coreCount'] = 1
    if options.site is not None:
        if ',' in options.site:
            taskParamMap['includedSite'] = PsubUtils.splitCommaConcatenatedItems([options.site])
        else:
            taskParamMap['site'] = options.site
    if options.evaluationContainer is not None:
        taskParamMap['container_name'] = options.evaluationContainer

    taskParamMap['multiStepExec'] = {'preprocess': {'command': '${TRF}',
                                                    'args': '--preprocess ${TRF_ARGS}'},
                                     'postprocess': {'command': '${TRF}',
                                                     'args': '--postprocess ${TRF_ARGS}'},
                                     'containerOptions': {'containerExec':
                                                              'while [ ! -f __payload_in_sync_file__ ]; do sleep 5; done; '
                                                              'echo "=== cat exec script ==="; '
                                                              'cat __run_main_exec.sh; '
                                                              'echo; '
                                                              'echo "=== exec script ==="; '
                                                              '/bin/sh __run_main_exec.sh; '
                                                              'REAL_MAIN_RET_CODE=$?; '
                                                              'touch __payload_out_sync_file__; '
                                                              'exit $REAL_MAIN_RET_CODE '
                                         ,
                                                          'containerImage': options.evaluationContainer}
                                     }
    if options.checkPointToSave is not None:
        taskParamMap['multiStepExec']['coprocess'] = {'command': '${TRF}',
                                                      'args': '--coprocess ${TRF_ARGS}'}

    if options.alrbArgs is not None:
        taskParamMap['multiStepExec']['containerOptions']['execArgs'] = options.alrbArgs

    logDatasetName = re.sub('/$', '.log/', options.outDS)

    taskParamMap['log'] = {'dataset': logDatasetName,
                           'container': logDatasetName,
                           'type': 'template',
                           'param_type': 'log',
                           'value': '{0}.$JEDITASKID.${{SN}}.log.tgz'.format(logDatasetName[:-1])
                           }

    taskParamMap['hpoRequestData'] = {'sandbox': options.steeringContainer,
                                      'executable': 'docker',
                                      'arguments': options.steeringExec,
                                      'output_json': 'output.json',
                                      'max_points': options.maxPoints,
                                      'num_points_per_generation': options.nPointsPerIteration,
                                      }
    if options.minUnevaluatedPoints is not None:
        taskParamMap['hpoRequestData']['min_unevaluated_points'] = options.minUnevaluatedPoints

    if options.searchSpaceFile is not None:
        with open(options.searchSpaceFile) as json_file:
            taskParamMap['hpoRequestData']['opt_space'] = json.load(json_file)

    taskParamMap['jobParameters'] = [
        {'type': 'constant',
         'value': '-o {0} -j "" --inSampleFile {1}'.format(options.evaluationOutput,
                                                           options.evaluationInput)
         },
        {'type': 'constant',
         'value': '-a {0} --sourceURL {1}'.format(archiveName, sourceURL)
         },
    ]

    taskParamMap['jobParameters'] += [
        {'type': 'constant',
         'value': '-p "',
         'padding': False,
         },
    ]
    taskParamMap['jobParameters'] += PsubUtils.convertParamStrToJediParam(options.evaluationExec, {}, '',
                                                                          True, False,
                                                                          includeIO=False)

    taskParamMap['jobParameters'] += [
        {'type': 'constant',
         'value': '"',
         },
    ]

    if options.checkPointToSave is not None:
        taskParamMap['jobParameters'] += [
            {'type': 'constant',
             'value': '--checkPointToSave {0}'.format(options.checkPointToSave)
             },
        ]
        if options.checkPointInterval is not None:
            taskParamMap['jobParameters'] += [
                {'type': 'constant',
                 'value': '--checkPointInterval {0}'.format(options.checkPointInterval)
                 },
            ]

    if options.checkPointToLoad is not None:
        taskParamMap['jobParameters'] += [
            {'type': 'constant',
             'value': '--checkPointToLoad {0}'.format(options.checkPointToLoad)
             },
        ]

    if options.trainingDS is not None:
        taskParamMap['jobParameters'] += [
            {'type': 'constant',
             'value': '--writeInputToTxt IN_DATA:{0}'.format(options.evaluationTrainingData)
             },
            {'type': 'template',
             'param_type': 'input',
             'value': '-i "${IN_DATA/T}"',
             'dataset': options.trainingDS,
             'attributes': 'nosplit,repeat',
             },
            {'type': 'constant',
             'value': '--inMap "{\'IN_DATA\': ${IN_DATA/T}}"'
             },
        ]

    if options.evaluationMeta is not None:
        taskParamMap['jobParameters'] += [
            {'type': 'constant',
             'value': '--outMetaFile={0}'.format(options.evaluationMeta),
             },
        ]

    if options.segmentSpecFile is not None:
        taskParamMap['segmentedWork'] = True

        with open(options.segmentSpecFile) as f:
            # read segments
            segments = json.load(f)
            # search space
            if 'opt_space' in taskParamMap['hpoRequestData'] and \
                    isinstance(taskParamMap['hpoRequestData']['opt_space'], dict):
                space = taskParamMap['hpoRequestData']['opt_space']
                taskParamMap['hpoRequestData']['opt_space'] = []
            else:
                space = None
            # set model ID to each segment
            for i in range(len(segments)):
                segments[i].update({'id': i})
                # make clone of search space if needed
                if space is not None:
                    new_space = dict()
                    new_space['model_id'] = i
                    new_space['search_space'] = copy.deepcopy(space)
                    taskParamMap['hpoRequestData']['opt_space'].append(new_space)
            taskParamMap['segmentSpecs'] = segments
            # multiply by num of segments
            taskParamMap['maxNumJobs'] *= len(segments)
            taskParamMap['totNumJobs'] *= len(segments)
            taskParamMap['hpoRequestData']['max_points'] *= len(segments)

        taskParamMap['jobParameters'] += [
            {'type': 'constant',
             'value': '--segmentID=${SEGMENT_ID}',
             },
        ]

    if options.evaluationMetrics is not None:
        lfn = '$JEDITASKID.metrics.${SN}.tgz'
        if options.segmentSpecFile is not None:
            lfn = '${MIDDLENAME}.' + lfn
        taskParamMap['jobParameters'] += [
            {'type': 'template',
             'param_type': 'output',
             'value': lfn,
             'dataset': options.outDS,
             'hidden': True,
             'allowNoOutput': True,
             },
            {'type': 'constant',
             'value': '--outMetricsFile=${{OUTPUT0}}^{0}'.format(options.evaluationMetrics),
             },
        ]

    if options.maxPointsPerEvaluationJob:
        taskParamMap['jobParameters'] += [
            {'type': 'constant',
             'value': '--maxLoopCount={}'.format(options.maxPointsPerEvaluationJob),
             },
        ]

    if options.noSubmit:
        if options.noSubmit:
            if options.verbose:
                tmpLog.debug("==== taskParams ====")
                tmpKeys = list(taskParamMap)
                tmpKeys.sort()
                for tmpKey in tmpKeys:
                    print('%s : %s' % (tmpKey, taskParamMap[tmpKey]))
        sys.exit(0)

    if get_taskparams:
        return taskParamMap

    tmpLog.info("submit {0}".format(options.outDS))
    tmpStat, tmpOut = Client.insertTaskParams(taskParamMap, options.verbose, True)
    # result
    taskID = None
    exitCode = None
    if tmpStat != 0:
        tmpStr = "task submission failed with {0}".format(tmpStat)
        tmpLog.error(tmpStr)
        exitCode = 1
    else:
        if tmpOut[0] in [0, 3]:
            tmpStr = tmpOut[1]
            tmpLog.info(tmpStr)
            try:
                m = re.search('jediTaskID=(\d+)', tmpStr)
                taskID = int(m.group(1))
            except Exception:
                pass
        else:
            tmpStr = "task submission failed. {0}".format(tmpOut[1])
            tmpLog.error(tmpStr)
            exitCode = 1

    dumpItem = copy.deepcopy(vars(options))
    dumpItem['returnCode'] = exitCode
    dumpItem['returnOut'] = tmpStr
    dumpItem['jediTaskID'] = taskID

    # dump
    if options.dumpJson is not None:
        with open(os.path.expanduser(options.dumpJson), 'w') as f:
            json.dump(dumpItem, f)



import os
import sys
import logging


rootLog = None

# set logger
def setLogger(tmpLog):
    global rootLog
    rootLog = tmpLog


# return logger
def getPandaLogger(use_stdout=True):
    # use root logger
    global rootLog
    if rootLog is None:
        rootLog = logging.getLogger('panda-client')
    # add StreamHandler if no handler
    if rootLog.handlers == []:
        rootLog.setLevel(logging.DEBUG)
        if use_stdout:
            console = logging.StreamHandler(sys.stdout)
        else:
            console = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter('%(levelname)s : %(message)s')
        console.setFormatter(formatter)
        rootLog.addHandler(console)
    # return
    return rootLog


# disable logging
def disable_logging():
    global rootLog
    if not rootLog:
        rootLog = logging.getLogger('')
    rootLog.disabled = True
    # keep orignal stdout mainly for jupyter
    sys.__stdout__ = sys.stdout
    sys.stdout = open(os.devnull, 'w')


# enable logging
def enable_logging():
    global rootLog
    if rootLog:
        rootLog.disabled = False
    sys.stdout.close()
    sys.stdout = sys.__stdout__



import os
import time
import datetime
import json
import sys
import ssl
import re

try:
    from urllib.parse import urlencode
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError, URLError
except ImportError:
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError, URLError

try:
    baseMonURL = os.environ['PANDAMON_URL']
except Exception:
    baseMonURL = 'https://bigpanda.cern.ch'

HEADERS = {'Accept': 'application/json', 'Content-Type':'application/json'}


def query_tasks(jeditaskid=None, username=None, limit=10000, taskname=None, status=None, superstatus=None,
                reqid=None, days=None, metadata=False, sync=False, verbose=False):
    timestamp = int(time.time())
    parmas = {  'json': 1,
                'datasets': True,
                'limit':limit,
                }
    if jeditaskid:
        parmas['jeditaskid'] = jeditaskid
    if username:
        parmas['username'] = username
    if taskname:
        parmas['taskname'] = taskname
    if status:
        parmas['status'] = status
    if superstatus:
        parmas['superstatus'] = superstatus
    if reqid:
        parmas['reqid'] = reqid
    if days is not None:
        parmas['days'] = days
    if metadata:
        parmas['extra'] = 'metastruct'
    if sync:
        parmas['timestamp'] = timestamp
    url = baseMonURL + '/tasks/?{0}'.format(urlencode(parmas))
    if verbose:
        sys.stderr.write('query url = {0}\n'.format(url))
        sys.stderr.write('headers   = {0}\n'.format(json.dumps(HEADERS)))
    try:
        req = Request(url, headers=HEADERS)
        try:
            # Skip SSL verification
            context = ssl._create_unverified_context()
        except AttributeError:
            # Legacy Python that doesn't verify HTTPS certificates by default
            rep = urlopen(req)
        else:
            rep = urlopen(req, context=context)
        if verbose:
            sys.stderr.write('time UTC  = {0}\n'.format(datetime.datetime.utcnow()))
        rec = rep.getcode()
        if verbose:
            sys.stderr.write('resp code = {0}\n'.format(rec))
        res = rep.read().decode('utf-8')
        if verbose:
            sys.stderr.write('data = {0}\n'.format(res))
        ret = json.loads(res)
        return timestamp, url, ret
    except Exception as e:
        err_str = '{0} : {1}'.format(e.__class__.__name__, e)
        sys.stderr.write('{0}\n'.format(err_str))
        raise


def datetime_parser(d):
    for k, v in d.items():
        if isinstance(v, str) and re.search("^\d{4}-\d{2}-\d{2}(\W|T)\d{2}:\d{2}:\d{2}", v):
            try:
                d[k] = datetime.datetime.strptime(v, '%Y-%m-%dT%H:%M:%S')
            except Exception:
                try:
                    d[k] = datetime.datetime.strptime(v, '%Y-%m-%d %H:%M:%S')
                except Exception:
                    pass
    return d


def query_jobs(jeditaskid, limit=10000, drop=True, verbose=False):
    parmas = {'json': 1,
              'limit':limit,
              }
    parmas['jeditaskid'] = jeditaskid
    if not drop:
        parmas['mode'] = 'nodrop'
    url = baseMonURL + '/jobs/?{0}'.format(urlencode(parmas))
    if verbose:
        sys.stderr.write('query url = {0}\n'.format(url))
        sys.stderr.write('headers   = {0}\n'.format(json.dumps(HEADERS)))
    try:
        req = Request(url, headers=HEADERS)
        try:
            # Skip SSL verification
            context = ssl._create_unverified_context()
        except AttributeError:
            # Legacy Python that doesn't verify HTTPS certificates by default
            rep = urlopen(req)
        else:
            rep = urlopen(req, context=context)
        if verbose:
            sys.stderr.write('time UTC  = {0}\n'.format(datetime.datetime.utcnow()))
        rec = rep.getcode()
        if verbose:
            sys.stderr.write('resp code = {0}\n'.format(rec))
        res = rep.read().decode('utf-8')
        if verbose:
            sys.stderr.write('data = {0}\n'.format(res))
        ret = json.loads(res, object_hook=datetime_parser)
        return url, ret
    except Exception as e:
        err_str = '{0} : {1}'.format(e.__class__.__name__, e)
        sys.stderr.write('{0}\n'.format(err_str))
        raise


"""
Do NOT import this module in your code.
Import PBookCore instead.
"""

import os
import sys
import code
import atexit
import signal
import tempfile

from pandaclient.MiscUtils import commands_get_output
try:
    long()
except Exception:
    long = int

try:
    from concurrent.futures import ThreadPoolExecutor
except ImportError:
    def list_parallel_exec(func, array):
        return [ func(x) for x in array ]
else:
    def list_parallel_exec(func, array):
        with ThreadPoolExecutor(8) as thread_pool:
            dataIterator = thread_pool.map(func, array)
        return list(dataIterator)


import argparse
import readline

from pandaclient import Client
from pandaclient import PandaToolsPkgInfo

# readline support
readline.parse_and_bind('tab: complete')
readline.parse_and_bind('set show-all-if-ambiguous On')

# history support
pconfDir = os.path.expanduser(os.environ['PANDA_CONFIG_ROOT'])
if not os.path.exists(pconfDir):
    os.makedirs(pconfDir)
historyFile = '%s/.history' % pconfDir
# history file
if os.path.exists(historyFile):
    try:
        # except for macOS X
        readline.read_history_file(historyFile)
    except Exception:
        pass
readline.set_history_length(1024)

# set dummy CMTSITE
if 'CMTSITE' not in os.environ:
    os.environ['CMTSITE'] = ''

# make tmp dir
tmpDir = tempfile.mkdtemp()

# set tmp dir in Client
Client.setGlobalTmpDir(tmpDir)

# fork PID
fork_child_pid = None

# exit action
def _onExit(dirName,hFile):
    # save history only for master process
    if fork_child_pid == 0:
        readline.write_history_file(hFile)
    # remove tmp dir
    commands_get_output('rm -rf %s' % dirName)
atexit.register(_onExit,tmpDir,historyFile)


# look for PandaTools package
for path in sys.path:
    if path == '':
        path = '.'
    if os.path.exists(path) and os.path.isdir(path) and 'pandaclient' in os.listdir(path) \
           and os.path.exists('%s/pandaclient/__init__.py' % path):
        # make symlink for module name
        os.symlink('%s/pandaclient' % path,'%s/taskbuffer' % tmpDir)
        break
sys.path = [tmpDir]+sys.path

from pandaclient import PBookCore    # noqa: E402

orig_help = help


# main for interactive session
def intmain(pbookCore,comString):

    # help
    def help(*arg):
        """
        Show the help doc
        """
        if len(arg) > 0:
            try:
                if isinstance(arg[0], str):
                    func = main_locals[arg[0]]
                else:
                    func = arg[0]
                orig_help(func)
                return
            except Exception:
                print("Unknown command : {0}".format(str(arg[0])))
        # print available methods
        tmp_str = """
The following commands are available:

    help
    show
    showl
    kill
    retry
    finish
    kill_and_retry
    get_user_job_metadata
    recover_lost_files
    show_workflow
    kill_workflow
    retry_workflow
    finish_workflow
    pause_workflow
    resume_workflow
    set_secret
    list_secrets
    delete_secret
    delete_all_secrets

For more info, do help(show) for example
"""
        print(tmp_str)

    # show status
    def show(*args, **kwargs):
        """
        Print task records. The first argument (non-keyword) can be an jediTaskID or reqID, or 'run' (show active tasks only), or 'fin' (show terminated tasks only), or can be omitted. The following keyword arguments are available in the way of panda monitor url query: [username, limit, taskname, days, jeditaskid].
        If sync=True, it forces panda monitor to get the latest records rather than get from cache.
        Specify display format with format='xxx', available formats are ['standard', 'long', 'json', 'plain'].
        The default filter conditions are: username=(name from user voms proxy), limit=1000, days=14, sync=False, format='standard'.

        example:
        >>> show()
        >>> show(123)
        >>> show(12345678, format='long')
        >>> show(taskname='my_task_name')
        >>> show('run')
        >>> show('fin', days=7, limit=100)
        >>> show(format='json', sync=True)
        """
        return pbookCore.show(*args, **kwargs)

    # shortcut to show long status
    def showl(*args, **kwargs):
        """
        Print task records in long format; shortcut function of show(..., format='long'). See help message of show() for other keywords arguments

        example:
        >>> showl()
        >>> showl(123)
        >>> showl(12345678)
        >>> showl(taskname='my_task_name')
        """
        kwargs['format'] = 'long'
        return pbookCore.show(*args, **kwargs)

    # kill
    def kill(taskIDs):
        """
        Kill all subJobs in taskIDs (ID or a list of ID, can be either jediTaskID or reqID). If 'all', kill all active tasks of the user.

         example:
           >>> kill(123)
           >>> kill([123, 345, 567])
           >>> kill('all')
        """
        if taskIDs == 'all':
            # active tasks
            task_list = pbookCore.get_active_tasks()
            ret = list_parallel_exec(lambda task: pbookCore.kill(task.jeditaskid), task_list)
        elif isinstance(taskIDs, (list, tuple)):
            ret = list_parallel_exec(lambda taskID: pbookCore.kill(taskID), taskIDs)
        elif isinstance(taskIDs, (int, long)):
            ret = [ pbookCore.kill(taskIDs) ]
        else:
            print('Error: Invalid argument')
            ret = None
        return ret

    # finish
    def finish(taskIDs, soft=False):
        """
        Finish all subJobs in taskIDs (ID or a list of ID, can be either jediTaskID or reqID). If taskIDs is 'all',
        finish all active tasks of the user. If soft is False (default), all running jobs are killed and the task
        finishes immediately. If soft is True, new jobs are not generated and the task finishes once all running
        jobs finish.

         example:
           >>> finish(123)
           >>> finish(234, soft=True)
           >>> finish([123, 345, 567])
           >>> finish('all')
        """
        if taskIDs == 'all':
            # active tasks
            task_list = pbookCore.get_active_tasks()
            ret = list_parallel_exec(lambda task: pbookCore.finish.original_func(pbookCore,
                                                                                 task.jeditaskid, soft=soft),
                                     task_list)
        elif isinstance(taskIDs, (list, tuple)):
            ret = list_parallel_exec(lambda taskID: pbookCore.finish(taskID, soft=soft), taskIDs)
        elif isinstance(taskIDs, (int, long)):
            ret = [ pbookCore.finish(taskIDs, soft=soft) ]
        else:
            print('Error: Invalid argument')
            ret = None
        return ret

    # retry
    def retry(taskIDs, newOpts=None, days=14, limit=1000, **kwargs):
        """
        Retry failed/cancelled subJobs in taskIDs (ID or a list of ID, can be either jediTaskID or reqID).
        It is possible to specify newOpts, which is a map of options and new arguments like
        {'nFilesPerJob':10,'excludedSite':'ABC,XYZ'}, to overwrite task parameters for new attempts. The list of
        changeable parameters is site, excludedSite, includedSite, nFilesPerJob, nGBPerJob, nFiles, nEvents,
        loopingCheck, nMaxFilesPerJob, ramCount, avoidVP. It is also possible to specify those parameters as named
        arguments of the retry function, e.g. nFilesPerJob=10, excludedSite='ABC,XYZ'.
        If input files were used or are being used by other jobs for the same
        output dataset container, those file are skipped to avoid job duplication when retrying failed subjobs.

        If taskIDs is 'all', it retries 1000 tasks at most that have finished for the last 14 days. It is possible
        to retry more tasks by setting the days and limit options. If newOpts and/or named arguments are specified,
        they are applied to all retried tasks.

         example:
           >>> retry(123)
           >>> retry([123, 345, 567])
           >>> retry(789, newOpts={'excludedSite':'siteA,siteB'})
           >>> retry(789, excludedSite='siteA,siteB')
           >>> retry('all')
           >>> retry('all', days=30, limit=2000)
           >>> retry('all', newOpts={'excludedSite':'siteA,siteB'})
        """
        if newOpts is None:
            newOpts = kwargs
        if isinstance(taskIDs, (list, tuple)):
            ret = list_parallel_exec(lambda taskID: pbookCore.retry(taskID, newOpts=newOpts), taskIDs)
        elif isinstance(taskIDs, (int, long)):
            ret = [pbookCore.retry(taskIDs, newOpts=newOpts)]
        elif taskIDs == 'all':
            dataList = pbookCore.show(status='finished', days=days, limit=limit, sync=True, format='json')
            ret = list_parallel_exec(lambda data: pbookCore.retry.original_func(pbookCore, data['jeditaskid'],
                                                                                newOpts=newOpts),
                                     dataList)
        else:
            print('Error: Invalid argument')
            ret = None
        return ret

    # debug mode
    def debug(PandaID, modeOn):
        """
        Turn the debug mode on/off for a subjob with PandaID. modeOn is True/False to enable/disable the debug mode. Note that the maxinum number of debug subjobs is limited. If you already hit the limit you need to disable the debug mode for a subjob before debugging another subjob

         example:
           >>> debug(1234, True)
        """
        pbookCore.debug(PandaID, modeOn)

    # kill and retry
    def killAndRetry(taskIDs, newOpts=None):
        if newOpts is None:
            newOpts = {}
        if isinstance(taskIDs, (list, tuple)):
            ret = list_parallel_exec(lambda taskID: pbookCore.killAndRetry(taskID, newOpts=newOpts), taskIDs)
        elif isinstance(taskIDs, (int, long)):
            ret = [ pbookCore.killAndRetry(taskIDs, newOpts=newOpts) ]
        else:
            print('Error: Invalid argument')
            ret = None
        return ret

    # kill and retry
    def kill_and_retry(taskIDs, newOpts=None, **kwargs):
        """
        Kill running sub-jobs and then retry failed/cancelled sub-jobs in taskIDs (ID or a list of ID, can be either jediTaskID or reqID). Concerning newOpts, see help(retry)

         example:
           >>> kill_and_retry(123)
           >>> kill_and_retry([123, 345, 567])
           >>> kill_and_retry(789, newOpts={'excludedSite':'siteA,siteB'})
           >>> kill_and_retry(789, excludedSite='siteA,siteB')
        """
        if newOpts is None:
            newOpts = kwargs
        return killAndRetry(taskIDs, newOpts)

    # get user job metadata
    def getUserJobMetadata(taskID, outputFileName):
        pbookCore.getUserJobMetadata(taskID, outputFileName)

    # get user job metadata
    def get_user_job_metadata(taskID, outputFileName):
        """
        Get user metadata of successful jobs in a task and write them in a json file

         example:
           >>> getUserJobMetadata(123, 'output.json')
        """
        getUserJobMetadata(taskID, outputFileName)

    # recover lost files
    def recover_lost_files(taskID, test_mode=False):
        """
        Send a request to recove lost files produced by a task. Set test_mode=True for testing

        example:
          >>> recover_lost_files(123)
          >>> recover_lost_files(123, test_mode=True)
        """
        pbookCore.recover_lost_files(taskID, test_mode)

    # finish a workflow
    def finish_workflow(request_id):
        """
        Send a request to finish a workflow

        """
        status, output = pbookCore.execute_workflow_command('finish', request_id)
        if output:
            print(output[0][-1])

    # kill a workflow
    def kill_workflow(request_id):
        """
        Send a request to kill a workflow

        """
        status, output = pbookCore.execute_workflow_command('abort', request_id)
        if output:
            print(output[0][-1])

    # pause a workflow
    def pause_workflow(request_id):
        """
        Send a request to pause a workflow

        """
        status, output = pbookCore.execute_workflow_command('suspend', request_id)
        if output:
            print(output[0][-1])

    # resume a workflow
    def resume_workflow(request_id):
        """
        Send a request to resume a workflow

        """
        status, output = pbookCore.execute_workflow_command('resume', request_id)
        if output:
            print(output[0][-1])

    # retry a workflow
    def retry_workflow(request_id):
        """
        Send a request to retry a workflow

        """
        status, output = pbookCore.execute_workflow_command('retry', request_id)
        if output:
            print(output[0][-1])

    # show a workflow
    def show_workflow(request_id):
        """
        Show a workflow

        """
        status, output = pbookCore.execute_workflow_command('get_status', request_id)
        if output:
            print(output)

    # set a secret
    def set_secret(key, value, is_file=False):
        """
        Define a pair of secret key-value strings. The value can be a file path to upload a secret file when is_file=True

        """
        pbookCore.set_secret(key, value, is_file)

    # delete a secret
    def delete_secret(key):
        """
        Delete a secret

        """
        pbookCore.set_secret(key, None)

    # delete all secrets
    def delete_all_secrets():
        """
        Delete all secrets

        """
        pbookCore.set_secret(None, None)

    # list secrets
    def list_secrets(full=False):
        """
        List secrets. Value strings are truncated by default. full=True to see entire strings

        """
        pbookCore.list_secrets(full)

    # execute command in the batch mode
    if comString != '':
        exec(comString) in globals(), locals()
        # exit
        if PBookCore.func_return_value:
            sys.exit(0)
        else:
            sys.exit(1)
    main_locals = locals()
    # go to interactive prompt
    code.interact(banner="\nStart pBook %s" % PandaToolsPkgInfo.release_version,
                  local=locals())


# kill whole process
def catch_sig(sig, frame):
    # cleanup
    _onExit(tmpDir,historyFile)
    # kill
    commands_get_output('kill -9 -- -%s' % os.getpgrp())


# overall main
def main():
    # parse option
    parser = argparse.ArgumentParser(conflict_handler="resolve")
    parser.add_argument("-v",action="store_true",dest="verbose",default=False,
                      help="Verbose")
    parser.add_argument('-c',action='store',dest='comString',default='',type=str,
                      help='Execute a command in the batch mode')
    parser.add_argument("-3", action="store_true", dest="python3", default=False,
                      help="Use python3")
    parser.add_argument('--version',action='store_const',const=True,dest='version',default=False,
                      help='Displays version')
    parser.add_argument('--devSrv',action='store_const',const=True,dest='devSrv',default=False,
                      help=argparse.SUPPRESS)
    parser.add_argument('--intrSrv',action='store_const',const=True, dest='intrSrv',default=False,
                      help=argparse.SUPPRESS)
    # option for jupyter notebook
    parser.add_argument('--prompt_with_newline', action='store_const', const=True, dest='prompt_with_newline',
                        default=False, help=argparse.SUPPRESS)

    options,args = parser.parse_known_args()

    # display version
    if options.version:
        print("Version: %s" % PandaToolsPkgInfo.release_version)
        sys.exit(0)

    # use dev server
    if options.devSrv:
        Client.useDevServer()

    # use INTR server
    if options.intrSrv:
        Client.useIntrServer()

    # fork for Ctl-c
    global fork_child_pid
    fork_child_pid = os.fork()
    if fork_child_pid == -1:
        print("ERROR : Failed to fork")
        sys.exit(1)
    if fork_child_pid == 0:
        # main
        # instantiate core
        if options.verbose:
            print(options)
        if options.prompt_with_newline:
            sys.ps1 = ">>> \n"
        pbookCore = PBookCore.PBookCore(verbose=options.verbose)

        # CUI
        intmain(pbookCore,options.comString)
    else:
        # set handler
        signal.signal(signal.SIGINT, catch_sig)
        signal.signal(signal.SIGHUP, catch_sig)
        signal.signal(signal.SIGTERM,catch_sig)
        pid, status = os.wait()
        if os.WIFSIGNALED(status):
            sys.exit(-os.WTERMSIG(status))
        elif os.WIFEXITED(status):
            sys.exit(os.WEXITSTATUS(status))
        else:
            sys.exit(0)



import copy


task_active_superstatus_list = ['running', 'submitting', 'registered', 'ready']
task_final_superstatus_list = ['finished', 'failed', 'done', 'broken', 'aborted']


class LocalTaskSpec(object):

    _attributes_hidden = (
                        '_pandaserver',
                        '_timestamp',
                        '_sourceurl',
                        '_weburl',
                        '_fulldict',
                        )

    _attributes_direct = (
                        'jeditaskid',
                        'reqid',
                        'taskname',
                        'username',
                        'creationdate',
                        'modificationtime',
                        'superstatus',
                        'status',
                        )

    _attributes_dsinfo = (
                        'pctfinished',
                        'pctfailed',
                        'nfiles',
                        'nfilesfinished',
                        'nfilesfailed',
                        )

    __slots__ = _attributes_hidden + _attributes_direct + _attributes_dsinfo

    # stdout string format
    strf_dict = {}
    strf_dict['standard'] = '{jtid:>10}  {reqid:>8}  {st:>10}  {pctf:>5}  {tname}'
    strf_dict['long'] = (   '{jtid:>10}  {st:>10}  {cret:20}  {modt:20}  ({filesprog})\n'
                            '{reqid:>10}  {pctf:>10}  {tname}\n'
                            '                        {weburl}\n' ) + '_'*78

    # header row
    head_dict = {}
    head_dict['standard'] = strf_dict['standard'].format(st='Status', jtid='JediTaskID',
                                                        reqid='ReqID', pctf='Fin%', tname='TaskName') \
                            + '\n' + '_'*64
    head_dict['long'] = strf_dict['long'].format(st='Status', jtid='JediTaskID', reqid='ReqID',
                                                tname='TaskName', weburl='Webpage', filesprog='finished|  failed|   total NInputFiles',
                                                pctf='Finished_%', cret='CreationDate', modt='ModificationTime')

    def __init__(self, task_dict, source_url=None, timestamp=None,
                    pandaserver='https://pandaserver.cern.ch:25443/server/panda'):
        self._timestamp = timestamp
        self._sourceurl = source_url
        self._pandaserver = pandaserver
        self._fulldict = copy.deepcopy(task_dict)
        for aname in self._attributes_direct:
            setattr(self, aname, self._fulldict.get(aname))
        for aname in self._attributes_dsinfo:
            if aname.startswith('pct'):
                setattr(self, aname, '{0}%'.format(self._fulldict['dsinfo'][aname]))
            else:
                setattr(self, aname, '{0}'.format(self._fulldict['dsinfo'][aname]))
        self._weburl = 'https://bigpanda.cern.ch/tasknew/{0}/'.format(self.jeditaskid)

    def is_terminated(self):
        if self.superstatus in task_final_superstatus_list:
            return True
        else:
            return False

    def print_plain(self):
        print('_'*64)
        str_format = '{attr:18} : \t{value}'
        for aname in self.__slots__:
            if aname in ['_fulldict']:
                continue
            print(str_format.format(attr=aname, value= getattr(self, aname)))

    def print_long(self):
        str_format = self.strf_dict['long']
        print(str_format.format(sst=self.superstatus, st=self.status, jtid=self.jeditaskid, reqid=self.reqid,
                                tname=self.taskname, weburl=self._weburl, pctf=self.pctfinished,
                                cret=self.creationdate, modt=self.modificationtime,
                                filesprog='{nff:>8}|{nfb:>8}|{nf:>8}'.format(
                                            nf=self.nfiles, nff=self.nfilesfinished, nfb=self.nfilesfailed)
                                ))

    def print_standard(self):
        str_format = self.strf_dict['standard']
        print(str_format.format(sst=self.superstatus, st=self.status, jtid=self.jeditaskid,
                                reqid=self.reqid, pctf=self.pctfinished, tname=self.taskname))



import re
import os
import json
import uuid
import traceback
import subprocess
try:
    import cPickle as pickle
except ImportError:
    import pickle
try:
    raw_input
except NameError:
    raw_input = input
try:
    unicode
except Exception:
    unicode = str


# wrapper for uuidgen
def wrappedUuidGen():
    return str(uuid.uuid4())


# make JEDI job parameter
def makeJediJobParam(lfn,dataset,paramType,padding=True,hidden=False,expand=False,
                     include='',exclude='',nFilesPerJob=None,offset=0,destination='',
                     token='',useNumFilesAsRatio=False,randomAtt=False,reusableAtt=False,
                     allowNoOutput=None, outDS=None, file_list=None):
    dictItem = {}
    if paramType == 'output':
        dictItem['type']       = 'template'
        dictItem['value']      = lfn
        dictItem['param_type'] = paramType
        dictItem['dataset']    = dataset
        dictItem['container']  = dataset
        if destination != '':
            dictItem['destination'] = destination
        if token != '':
            dictItem['token'] = token
        if not padding:
            dictItem['padding'] = padding
        if allowNoOutput is not None:
            for tmpPatt in allowNoOutput:
                if tmpPatt == '':
                    continue
                tmpPatt = '^.*'+tmpPatt+'$'
                if re.search(tmpPatt,lfn) is not None:
                    dictItem['allowNoOutput'] = True
                    break
    elif paramType == 'input':
        dictItem['type']       = 'template'
        dictItem['value']      = lfn
        dictItem['param_type'] = paramType
        dictItem['dataset']    = dataset
        if offset > 0:
            dictItem['offset'] = offset
        if include != '':
            dictItem['include'] = include
        if exclude != '':
            dictItem['exclude'] = exclude
        if expand:
            dictItem['expand'] = expand
        elif outDS:
            dictItem['consolidate'] = '.'.join(outDS.split('.')[:2]) + '.' + wrappedUuidGen() + '/'
        if nFilesPerJob not in [None,0]:
            dictItem['nFilesPerJob'] = nFilesPerJob
        if useNumFilesAsRatio and nFilesPerJob not in [None,0]:
            dictItem['ratio'] = nFilesPerJob
        if file_list:
            dictItem['files'] = file_list
    if hidden:
        dictItem['hidden'] = hidden
    if randomAtt:
        dictItem['random'] = True
    if reusableAtt:
        dictItem['reusable'] = True
    return [dictItem]


# get dataset name and num of files for a stream
def getDatasetNameAndNumFiles(streamDS,nFilesPerJob,streamName):
    if streamDS == "":
        # read from stdin
        print("\nThis job uses %s stream" % streamName)
        while True:
            streamDS = raw_input("Enter dataset name for {0}: ".format(streamName))
            streamDS = streamDS.strip()
            if streamDS != "":
                break
    # number of files per one signal
    if nFilesPerJob < 0:
        while True:
            tmpStr = raw_input("Enter the number of %s files per job : " % streamName)
            try:
                nFilesPerJob = int(tmpStr)
                break
            except Exception:
                pass
    # return
    return streamDS,nFilesPerJob


# convert UTF-8 to ASCII in json dumps
def unicodeConvert(input):
    if isinstance(input, dict):
        retMap = {}
        for tmpKey in input:
            tmpVal = input[tmpKey]
            retMap[unicodeConvert(tmpKey)] = unicodeConvert(tmpVal)
        return retMap
    elif isinstance(input, list):
        retList = []
        for tmpItem in input:
            retList.append(unicodeConvert(tmpItem))
        return retList
    elif isinstance(input, unicode):
        return input.encode('ascii', 'ignore').decode()
    return input


# decode json with ASCII
def decodeJSON(input_file):
    with open(input_file) as f:
        return json.load(f, object_hook=unicodeConvert)


# replacement for commands
def commands_get_status_output(com):
    data = ''
    try:
        # for python 2.6
        #data = subprocess.check_output(com, shell=True, universal_newlines=True, stderr=subprocess.STDOUT)
        p = subprocess.Popen(com, shell=True, universal_newlines=True, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        data, unused_err = p.communicate()
        retcode = p.poll()
        if retcode:
            ex = subprocess.CalledProcessError(retcode, com)
            raise ex
        status = 0
    except subprocess.CalledProcessError as ex:
        # for python 2.6
        #data = ex.output
        status = ex.returncode
    if data[-1:] == '\n':
        data = data[:-1]
    return status, data


def commands_get_output(com):
    return commands_get_status_output(com)[1]


# decorator to run with the original environment
def run_with_original_env(func):
    def new_func(*args, **kwargs):
        if 'LD_LIBRARY_PATH_ORIG' in os.environ and 'LD_LIBRARY_PATH' in os.environ:
            os.environ['LD_LIBRARY_PATH_RESERVE'] = os.environ['LD_LIBRARY_PATH']
            os.environ['LD_LIBRARY_PATH'] = os.environ['LD_LIBRARY_PATH_ORIG']
        if 'PYTHONPATH_ORIG' in os.environ:
            os.environ['PYTHONPATH_RESERVE'] = os.environ['PYTHONPATH']
            os.environ['PYTHONPATH'] = os.environ['PYTHONPATH_ORIG']
        if 'PYTHONHOME_ORIG' in os.environ and os.environ['PYTHONHOME_ORIG'] != '':
            if 'PYTHONHOME' in os.environ:
                os.environ['PYTHONHOME_RESERVE'] = os.environ['PYTHONHOME']
            os.environ['PYTHONHOME'] = os.environ['PYTHONHOME_ORIG']
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(str(e) + traceback.format_exc())
            raise e
        finally:
            if 'LD_LIBRARY_PATH_RESERVE' in os.environ:
                os.environ['LD_LIBRARY_PATH'] = os.environ['LD_LIBRARY_PATH_RESERVE']
            if 'PYTHONPATH_RESERVE' in os.environ:
                os.environ['PYTHONPATH'] = os.environ['PYTHONPATH_RESERVE']
            if 'PYTHONHOME_RESERVE' in os.environ:
                os.environ['PYTHONHOME'] = os.environ['PYTHONHOME_RESERVE']
    return new_func


# run commands with the original environment
@run_with_original_env
def commands_get_output_with_env(com):
    return commands_get_output(com)


@run_with_original_env
def commands_get_status_output_with_env(com):
    return commands_get_status_output(com)


# unpickle python2 pickle with python3
def pickle_loads(str_input):
    try:
        return pickle.loads(str_input)
    except Exception:
        try:
            return pickle.loads(str_input.encode('utf-8'), encoding='latin1')
        except Exception:
            return str_input


# parse secondary dataset option
def parse_secondary_datasets_opt(secondaryDSs):
    if secondaryDSs != '':
        # parse
        tmpMap = {}
        for tmpItem in secondaryDSs.split(','):
            if "#" in tmpItem:
                tmpItems = tmpItem.split('#')
            else:
                tmpItems = tmpItem.split(':')
            if 3 <= len(tmpItems) <= 6:
                tmpDsName = tmpItems[2]
                # change ^ to ,
                tmpDsName = tmpDsName.replace('^',',')
                # make map
                tmpMap[tmpDsName] = {'nFiles'     : int(tmpItems[1]),
                                     'streamName' : tmpItems[0],
                                     'pattern'    : '',
                                     'nSkip'      : 0,
                                     'files'      : []}
                # using filtering pattern
                if len(tmpItems) >= 4 and tmpItems[3]:
                    tmpMap[tmpItems[2]]['pattern'] = tmpItems[3]
                # nSkip
                if len(tmpItems) >= 5 and tmpItems[4]:
                    tmpMap[tmpItems[2]]['nSkip'] = int(tmpItems[4])
                # files
                if len(tmpItems) >= 6 and tmpItems[5]:
                    with open(tmpItems[5]) as f:
                        for l in f:
                            l = l.strip()
                            if l:
                                tmpMap[tmpItems[2]]['files'].append(l)
            else:
                errStr = "Wrong format %s in --secondaryDSs. Must be "\
                         "StreamName:nFilesPerJob:DatasetName[:Pattern[:nSkipFiles[:FileNameList]]]" \
                             % tmpItem
                return False, errStr
        # set
        secondaryDSs = tmpMap
    else:
        secondaryDSs = {}
    return True, secondaryDSs